# Εισαγωγή στους υπολογιστές με τη γλώσσα Python. εκδ.4
# Κεφάλαιο 15 - Μόνιμη αποθήκευση δεδομένων
# εφαρμογή university - έκδοση 2 (pickle)
import random
import os
import pickle
dir = "."
class Student():
    '''κλάση φοιτητών - περιέχει και τις βαθμολογίες'''
    theStudents = {}
    @staticmethod
    def student_list():
        lista =""
        for am in sorted([x for x in Student.theStudents], key=lambda x: Student.theStudents[x].surname):
            s = Student.theStudents[am]
            average_grade = sum(s.courses.values())/len(s.courses)
            lista += "{}\t{:15s}\t{:15s}\t{:.1f}\n".format(s.am, s.surname, s.name, average_grade)
        return lista
    def __init__(self, am, name, surname):
        self.am = am
        self.name = name
        self.surname = surname
        self.courses = {}
        Student.theStudents[am] = self
    def record(self):
        rec = "\n[{}]\t{}  {}\n".format(self.am, self.surname, self.name)
        rec += 50*"_"+"\n"
        rec += "{:5s}\t{:30s}\t{:10s}\n".format("Κωδ.", "Μάθημα", "βαθμός")
        for c in self.courses:
            course = Course.theCourses[c]
            rec += "{:5s}\t{:30s}\t{:10s}\n".format(course.id, course.name, str(self.courses[c]))
        rec += "Μέση βαθμολογία = {:.2f}\n".format(sum(self.courses.values())/len(self.courses))
        return rec

class Course():
    '''κλάση μαθημάτων'''
    theCourses = {}
    def __init__(self, id, name):
        self.id = id
        self.name = name
        self.students = []
        Course.theCourses[id] = self
        self.average = None
    def course_list(self):
        grades = []
        lista = "\n{}\t[{}]{:20s}\n".format("ΜΑΘΗΜΑ:", self.id, self.name)
        lista += 60*"_"+"\n"
        for s in sorted([x for x in self.students], key=lambda x: x.surname):
            lista += "{}\t{:20s}\t{:20s}\t{:5s}\n".format(s.am, s.surname, s.name, str(s.courses[self.id]))
            grades.append(s.courses[self.id])
        if len(grades) > 0: self.average = sum(grades)/len(grades)
        lista += "Μέση βαθμολογία = {:.2f}\n".format(self.average)
        return lista

class Persistant():
    ''' κλάση μόνιμης αποθήκευσης των αντικειμένων Student, Course'''
    def __init__(self):
        self.s_file = "students.pickle"
        self.c_file = "courses.pickle"
    def store(self):
        with open(self.s_file, 'wb') as f:
            pickle.dump(Student.theStudents, f)
        with open(self.c_file, 'wb') as f:
            pickle.dump(Course.theCourses, f)
    def retrieve(self):
        if os.path.isfile(self.s_file):
            with open(self.s_file, 'rb') as f:
                Student.theStudents = pickle.load(f)
        else: return False
        if os.path.isfile(self.c_file):
            with open(self.c_file, 'rb') as f:
                Course.theCourses = pickle.load(f)
        else: return False
        return True

class Main():
    def __init__(self):
        persist = Persistant()
        if not persist.retrieve():
            print ("******* new class")
            self.create_class(size=10)
            persist.store()
        self.main_menu()
    def main_menu(self):
        while True:
            print("0. Μέση βαθμολογία φοιτητών\n1. Καρτέλα φοιτητή\n2. Βαθμολόγιο μαθήματος\n3. Στατιστικά Μαθημάτων\n<enter: exit>\n")
            sel = input()
            if not sel: break
            if sel == "0":
                print(Student.student_list())
            if sel == "1":
                selected = input("Κωδικός φοιτητή  ... 101-{}...".format(100+len(Student.theStudents)))
                if selected in Student.theStudents.keys():
                    s = Student.theStudents[selected]
                    print(s.record())
            if sel == "2":
                selected = input("Μάθημα: (επιλογές {}) ...".format(",".join(Course.theCourses.keys())))
                if selected in Course.theCourses.keys():
                    c = Course.theCourses[selected]
                    print(c.course_list())
            if sel == "3":
                for (id,c) in Course.theCourses.items():
                    if not c.average:
                        c.course_list()
                    print("{}\t{:30s}\t{:.2f} ({} φοιτ.)".format(c.id, c.name, c.average, len(c.students)))
                print()
    def create_class(self, size):
        names = []
        with open("names.txt", 'r', encoding='utf-8') as fin:
            for line in fin: names.append(line.strip())
        students = random.sample(names, size)
        with open("courses.txt", 'r', encoding='utf-8') as fin:
            for line in fin: Course(*line.strip().split(";")) # Aντικείμενο Course
        for i,x in enumerate(students):
            s = Student(am = str(100+i+1), name=x.split(";")[0], surname= x.split(";")[1]) # Αντικ. Student
            enrolled = random.sample(Course.theCourses.keys(), 5) # εγγραφή σε 5 μαθήματα
            for x in enrolled:
                s.courses[x] = random.randint(1,10) # βαθμός στο μάθημα
                Course.theCourses[x].students.append(s)

if __name__ == "__main__": Main()
