# Εισαγωγή στους υπολογιστές με τη γλώσσα Python. εκδ.4
# Κεφάλαιο 18 - Διαδικτυακές εφαρμογές
# openweathermap API
import datetime
import json
import urllib.request
import os.path

apikey = ""
# το κλειδί ανακτάται μετά από εγγραφή στην http://home.openweathermap.org/

def timestamp_to_time(ts):
    return (datetime.datetime.fromtimestamp(int(ts)).strftime('%H:%M:%S'))
def time_stamp_to_date(ts):
    return (datetime.datetime.fromtimestamp(int(ts)).strftime('%Y-%m-%d %H:%M:%S'))
def data_fetch(city_id):
    if not apikey:
        print("Παρακαλώ προμηθευτείτε ένα κλειδί από openweathermap.org")
        return {}
    else:
        api_url = "http://api.openweathermap.org/data/2.5/weather?id={}&mode=json&units=metric&APPID={}".format(str(city_id), apikey)
        print(api_url)
        url = urllib.request.urlopen(api_url)
        output = url.read().decode('utf-8')
        raw_api_dict = json.loads(output)
        url.close()
        return raw_api_dict
def extract_data(raw_api_dict):
    if not raw_api_dict: return {}
    data = dict(
        id = raw_api_dict.get('id'),
        city=raw_api_dict.get('name'),
        country=raw_api_dict.get('sys').get('country'),
        temp=raw_api_dict.get('main').get('temp'),
        humidity=raw_api_dict.get('main').get('humidity'),
        pressure=raw_api_dict.get('main').get('pressure'),
        sky=raw_api_dict['weather'][0]['main'],
        sunrise=timestamp_to_time(raw_api_dict.get('sys').get('sunrise')),
        sunset=timestamp_to_time(raw_api_dict.get('sys').get('sunset')),
        wind=raw_api_dict.get('wind').get('speed'),
        wind_deg=raw_api_dict.get('wind').get('deg'),
        dt=raw_api_dict.get('dt'),
        cloudiness=raw_api_dict.get('clouds').get('all'))
    return data
def display_and_save(data):
    if not data: return ""
    temp_symbol = '\xb0' + 'C' # βαθμοί κελσίου
    out = 60*'-'+'\n'
    out += 'Current weather in: {}, {}:\n'.format(data['city'], data['country'])
    out += '{} {},  {}\n'.format(data['temp'], temp_symbol, data['sky'])
    out += 'Wind Speed: {}, Degree: {}\n'.format(data['wind'], data['wind_deg'])
    out += 'Humidity: {}%\n'.format(data['humidity'])
    out += 'Cloud: {}%\n'.format(data['cloudiness'])
    out += 'Pressure: {}\n'.format(data['pressure'])
    out += 'Sunrise at: {} - Sunset at: {}\n'.format(data['sunrise'], data['sunset'])
    out += 'Last update from the server: {}\n'.format(timestamp_to_time(data['dt']))
    out += 60*'-'+'\n'
    filename = str(data['id']) + time_stamp_to_date(data['dt']).split()[0]
    with open(filename, 'w', encoding="utf-8")as f:
        f.write(out)
    return out
def check_if_exists(city_id):
    filename = str(city_id)+datetime.datetime.now().strftime("%Y-%m-%d")
    if os.path.isfile(filename):
        with open(filename, 'r', encoding="utf-8") as f:
            weather = f.read()
        return weather
    else: return False
def main():
    cities = []
    with open('cities.txt', 'r', encoding='utf-8') as f:
        for line in f:
            line = line.split()
            if len(line) == 3: cities.append(tuple(line[1:]))
    while True:
        print(','.join(str(i+1)+":"+city for i,city in enumerate([x[0] for x in cities])))
        city = input('Επιλογή πόλης (enter για έξοδο): ')
        if not city: break
        if city.isdigit() and 0 < int(city) <= len(cities)+1:
            city_id = cities[int(city)-1][1]
        existing_weather = check_if_exists(city_id)
        if existing_weather: print(existing_weather)
        else:
            data = data_fetch(city_id)
            if ('message' in data.keys()): print(data['message'])
            else: print(display_and_save(extract_data(data)))
if __name__ == '__main__':
    main()
